import argparse
import sys

from pathlib import Path

from d64.d64_image import D64Image
from d64.d71_image import D71Image
from d64.d81_image import D81Image


CREATE_METHODS = {
    'd64': D64Image.create,
    'd71': D71Image.create,
    'd81': D81Image.create
}


def main():
    parser = argparse.ArgumentParser(description='Create empty Commodore disk images.')
    parser.add_argument('label', help='disk label')
    parser.add_argument('id', help='disk identifier')
    parser.add_argument('filename', type=Path, help='image filename')
    parser.add_argument('--type', default='d64', choices=CREATE_METHODS.keys(), help='image type')
    args = parser.parse_args()

    if args.filename.exists():
        sys.exit("{!s} already exists".format(args.filename))
    print("Creating empty disk image as {!s}, {}:{}, type {}".format(args.filename, args.label, args.id, args.type))

    CREATE_METHODS[args.type](args.filename, args.label.encode(), args.id.encode())



from .block import Block
from .exceptions import DiskFullError
from .side_sector import SideSector


class File(object):
    """Read and write access to files."""
    def __init__(self, entry, mode):
        self.entry = entry
        if mode[0] == 'r':
            self.block = entry.first_block()
            self.read_pos = 2
        else:
            self.image = entry.block.image
            self.block = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, _, __):
        if exc_type is None:
            self.close()

    def get_first_block(self):
        """Return the first empty block for a file write."""
        next_block = self.image.alloc_first_block()
        if next_block is None:
            raise DiskFullError()
        next_block.data_size = 0
        self.entry.set_first_block(next_block)
        self.entry.size = 1
        return next_block

    def get_new_block(self):
        """Get a new empty block for a file write."""
        next_block = self.image.alloc_next_block(self.block.track, self.block.sector)
        if next_block is None:
            raise DiskFullError()
        next_block.data_size = 0
        self.block.set_next_block(next_block)
        self.entry.size += 1
        return next_block

    def read(self, count=-1):
        """Read bytes from file."""
        ret = b''

        while count:
            remaining = self.block.data_size-(self.read_pos-2)
            if remaining == 0:
                break

            # read as much as is wanted from the current block
            length = remaining if count == -1 else min(count, remaining)
            ret += self.block.get(self.read_pos, self.read_pos+length)
            self.read_pos += length
            if count != -1:
                count -= length

            if self.block.is_final:
                # no more blocks, end of file
                break

            if self.read_pos == Block.SECTOR_SIZE:
                # end of block, move on to the next block
                self.block = self.block.next_block()
                self.read_pos = 2

        return ret

    def write(self, data):
        """Write data to a file."""
        written = 0

        if self.block is None:
            # allocate first block
            self.block = self.get_first_block()

        while data:
            remaining_space = Block.SECTOR_SIZE-(self.block.data_size+2)
            if remaining_space:
                length = min(remaining_space, len(data))
                self.block.set(self.block.data_size+2, data[:length])
                self.block.data_size += length
                written += length
                data = data[length:]
            else:
                # end of block, append a new one
                self.block = self.get_new_block()

        return written

    def close(self):
        """Close file."""
        if self.block is None:
            # file open for write, no data written
            # DOS files cannot be empty, they must contain at least one byte
            self.write(b'\r')


class RelativeFile(File):
    """Read and write access to relative files."""
    def get_first_block(self):
        """Return the first empty block for a file write."""
        data_block = super().get_first_block()
        block = self.image.alloc_next_block(data_block.track, data_block.sector)
        side_sector = SideSector(self.image, block.track, block.sector)
        side_sector.number = 0
        side_sector.record_len = self.entry.record_len
        side_sector.add_side_sector(side_sector)
        side_sector.clear_data_blocks()
        side_sector.add_data_block(data_block)
        self.entry.side_sector_ts = (side_sector.track, side_sector.sector)
        self.side_sector = side_sector
        self.entry.size += 1
        return data_block

    def get_new_block(self):
        """Get a new empty block for a file write."""
        if len(self.side_sector.all_data_blocks()) == SideSector.MAX_DATA_LINKS:
            # current side sector is full, allocate a new one
            block = self.image.alloc_next_block(self.side_sector.track, self.side_sector.sector)
            side_sector = SideSector(self.image, block.track, block.sector)
            side_sector.number = self.side_sector.number+1
            side_sector.record_len = self.entry.record_len
            side_sector.set_peers(self.side_sector.all_side_sectors())
            side_sector.clear_data_blocks()
            self.side_sector.set_next_block(side_sector)
            self.side_sector = side_sector
            self.entry.size += 1

            # update the peer list of all side sectors (including the new one)
            ss = SideSector(self.image, *self.entry.side_sector_ts)
            while ss:
                ss.add_side_sector(side_sector)
                ss = ss.next_block()

        data_block = super().get_new_block()
        self.side_sector.add_data_block(data_block)
        return data_block

    def read_record(self):
        """Read a complete record from a file."""
        return self.read(self.entry.record_len)

    def write(self, data):
        """Write one or more records to a file."""
        ret = 0
        while data:
            chunk = data[:self.entry.record_len]
            ret += super().write(chunk.ljust(self.entry.record_len, b'\x00'))
            data = data[self.entry.record_len:]
        return ret

    def close(self):
        """Close file."""
        # pad rest of data block with empty records
        while self.block is None or Block.SECTOR_SIZE-(self.block.data_size+2) >= self.entry.record_len:
            self.write(b'\xff')



import argparse
import glob
import string
import sys

from pathlib import Path

import cmd2

from d64 import DiskImage


def is_image_path(p):
    return p[0] in string.digits and p[1] == ':'


class D64Shell(cmd2.Cmd):
    def __init__(self):
        super().__init__()

        self.encoding = 'ascii'
        self.add_settable(cmd2.Settable('encoding', str, 'Text encoding'))

        self.images = {}

    def expand_paths(self, files_in):
        ret = []
        for f in files_in:
            if is_image_path(f):
                if len(f) == 2:
                    # just a drive
                    ret.append(f)
                else:
                    image = self.images[int(f[0])]
                    ret += [f[:2]+n.name.decode(self.encoding) for n in image.glob(f[2:].encode(self.encoding))]
            else:
                ret += glob.glob(f)
        return ret

    attach_parser = argparse.ArgumentParser()
    attach_parser.add_argument('--drive', type=int, help="drive number")
    attach_parser.add_argument('--read-only', action='store_true', help="prevent modifications to image")
    attach_parser.add_argument('image', help="image file name")

    @cmd2.with_argparser(attach_parser)
    def do_attach(self, args):
        """Attach an image to a drive number."""
        if args.drive is None:
            # find first free drive
            for d in range(0, 10):
                if d not in self.images:
                    args.drive = d
                    break
            if args.drive is None:
                self.perror("All drives in use")
                return None
        elif args.drive in self.images:
            self.perror("Drive {} in use".format(args.drive))
            return None

        self.images[args.drive] = DiskImage(args.image).open()
        self.poutput("Attached {} to {}".format(args.image, args.drive))

    detach_parser = argparse.ArgumentParser()
    detach_parser.add_argument('drive', type=int, nargs='+', help="drive to detach")

    @cmd2.with_argparser(detach_parser)
    def do_detach(self, args):
        """Detach an image from a drive letter."""
        for drive in args.drive:
            if drive in self.images:
                self.images[drive].close()
                del self.images[drive]
                self.poutput("Detached {}".format(drive))
            else:
                self.perror("Invalid drive: {}". format(drive))

    def do_images(self, args):
        """Display the attached images."""
        if self.images:
            for d in sorted(self.images.keys()):
                self.poutput("    {}  {!s}".format(d, self.images[d]))
        else:
            self.poutput("No attached images")

    dir_parser = argparse.ArgumentParser()
    dir_parser.add_argument('drive', type=int, nargs='*', help="drive to list")

    @cmd2.with_argparser(dir_parser)
    def do_dir(self, args):
        """Display directory of a drive."""
        if args.drive:
            drive_list = args.drive
        else:
            drive_list = sorted(self.images.keys())

        first = True
        for d in drive_list:
            if d in self.images:
                if first:
                    first = False
                else:
                    self.poutput('')
                for l in self.images[d].directory(self.encoding, drive=d):
                    self.poutput(l)
            else:
                self.perror("Invalid drive: {}". format(d))

    copy_parser = argparse.ArgumentParser()
    copy_parser.add_argument('file', nargs=(2,), help="file source or destination")

    @cmd2.with_argparser(copy_parser)
    def do_copy(self, args):
        files = self.expand_paths(args.file)
        print(str(files))

        if is_image_path(files[-1]):
            # destination is an image
            dest = files[-1]
            if len(files) > 2 and len(dest) != 2:
                self.perror("Target '{}' is not a drive".format(dest))
                return None
            raise NotImplementedError("Destination is image")
        else:
            # destination is file system
            dest = Path(files[-1])
            if len(files) > 2:
                if not dest.is_dir():
                    self.perror("Target '{!s}' is not a directory".format(dest))
                    return None

    complete_attach = cmd2.Cmd.path_complete


def main():
    c = D64Shell()
    ret = c.cmdloop()

    # close any attached images
    for i in c.images.values():
        i.close()

    sys.exit(ret)



from .block import Block
from .d81_bam import D81BAM
from .d81_dir_entry import D81DirEntry
from .dos_image import DOSImage
from .partition import Partition


class D81Image(DOSImage):

    DIR_ENTRY = D81DirEntry
    DOS_VERSION = ord('3')
    DOS_FORMAT = ord('D')
    MAX_TRACK = 80
    DIR_TRACK = 40
    DIR_SECTOR = 3
    INTERLEAVE = 1
    TRACK_SECTOR_MAX = ((40, (1, 80)),)
    IMAGE_SIZES = (819200, 822400)

    def __init__(self, filename):
        self.bam = D81BAM(self)
        super().__init__(filename)

    def open(self, mode):
        super().open(mode)
        self.side_a_bam_block = Block(self, self.DIR_TRACK, 1)
        self.side_b_bam_block = Block(self, self.DIR_TRACK, 2)

    def partition(self, name):
        paths = [e for e in self.glob(name)]
        if paths:
            if paths[0].entry.file_type == 'CBM':
                # existing partition
                return Partition(self, entry=paths[0].entry)
            raise TypeError("Entry is not a partition")
        return Partition(self, name=name)

    def alloc_next_block(self, track, sector, directory=False):
        return self._alloc_next_block(track, sector, self.INTERLEAVE)

    @property
    def dos_version(self):
        return self.dir_block.get(0x19)

    @property
    def name(self):
        name = self.dir_block.get(4, 0x14)
        return name.rstrip(b'\xa0')

    @property
    def id(self):
        id = self.dir_block.get(0x16, 0x18)
        return id

    @property
    def dos_type(self):
        return self.dir_block.get(2)

    @dos_version.setter
    def dos_version(self, version):
        self.dir_block.set(0x19, version)

    @name.setter
    def name(self, nam):
        self.dir_block.set(4, nam[:16].ljust(16, b'\xa0'))

    @id.setter
    def id(self, did):
        if len(did) != 2:
            raise ValueError("Invalid disk id, "+str(did))
        self.dir_block.set(0x16, did)
        # update copy in BAM blocks
        self.side_a_bam_block.set(4, did)
        self.side_b_bam_block.set(4, did)

    @dos_type.setter
    def dos_type(self, dtype):
        self.dir_block.set(2, dtype)
        self.dir_block.set(0x1a, dtype)
        # update copy in BAM blocks
        self.side_a_bam_block.set(2, dtype)
        self.side_a_bam_block.set(3, dtype ^ 0xff)
        self.side_b_bam_block.set(2, dtype)
        self.side_b_bam_block.set(3, dtype ^ 0xff)

    @classmethod
    def create(cls, filepath, disk_name, disk_id):
        """Create an empty disk image."""
        super().create(filepath)

        image = cls(filepath)
        try:
            image.open('r+b')

            # block 40/0 contains various identifying fields
            header_block = Block(image, cls.DIR_TRACK, 0)
            header_block.set(4, b'\xa0' * 0x15)
            image.name = disk_name
            image.id = disk_id
            image.dos_type = cls.DOS_FORMAT
            image.dos_version = cls.DOS_VERSION

            # populate the BAM with all free blocks
            for sectors, range_ in cls.TRACK_SECTOR_MAX:
                for t in range(range_[0], range_[1]+1):
                    bits_free = '1' * sectors
                    image.bam.set_entry(t, sectors, bits_free)

            # block 40/3 contains 8 empty directory entries
            dir_block = Block(image, cls.DIR_TRACK, 3)
            dir_block.data_size = 0xfe

            # blocks 40/1 and 40/2 hold the BAM
            bam1_block = Block(image, cls.DIR_TRACK, 1)
            bam2_block = Block(image, cls.DIR_TRACK, 2)
            bam2_block.data_size = 0xfe

            # link BAM 1 block to BAM 2 block
            bam1_block.set_next_block(bam2_block)

            # link header block to directory block
            header_block.set_next_block(dir_block)

            # allocate all blocks
            for i in range(0, 4):
                image.bam.set_allocated(cls.DIR_TRACK, i)
        finally:
            image.close()



import argparse
import sys

from collections import defaultdict
from pathlib import Path

from d64 import DiskImage
from d64.d71_image import D71Image
from d64.d81_image import D81Image
from d64.block import Block
from d64.side_sector import SideSector


IMAGE = None
FIX = False
YES = False
VERBOSE = False
QUIET = False

USED_BLOCKS = defaultdict(set)


class CheckAborted(KeyboardInterrupt):
    pass


def remember_used(track, sector):
    """Note block usage for later reconciliation."""
    global USED_BLOCKS

    USED_BLOCKS[track].add(sector)


def fix_error(text, fixer=None, **kwargs):
    """Report, and optionally repair, an error."""
    print("ERROR: "+text)
    if FIX and fixer:
        if YES:
            return fixer(**kwargs)
        response = input("Fix? ")
        if response.lower() in ('q', 'quit'):
            raise CheckAborted
        if response.lower() in ('y', 'yes'):
            return fixer(**kwargs)
    return 1


def check_misc():
    """Check various DOS fields and report disk name & id."""
    errors = 0

    if not QUIET:
        print("\nChecking DOS information...")
    if IMAGE.dos_version != IMAGE.DOS_VERSION:
        msg = "Unknown DOS version, "+chr(IMAGE.dos_version)
        errors += fix_error(msg, fix_dos_version)
    elif VERBOSE:
        print("DOS version: "+chr(IMAGE.dos_version))

    if IMAGE.dos_type != IMAGE.DOS_FORMAT:
        msg = "Invalid DOS format, ${:02x}".format(IMAGE.dos_type)
        errors += fix_error(msg, fix_dos_types)
    elif VERBOSE:
        print("DOS format: "+chr(IMAGE.dos_type))

    block = Block(IMAGE, IMAGE.DIR_TRACK, 0)
    bad_link = True
    if not block.is_final:
        try:
            next = block.next_block()
            if next.track == IMAGE.DIR_TRACK and next.sector == IMAGE.DIR_SECTOR:
                bad_link = False
        except ValueError:
            # invalid track/sector
            pass
    if bad_link:
        ts = block.get(0, 2)
        msg = "Invalid link to directory block, {:d}:{:d}".format(*ts)
        errors += fix_error(msg, fix_dir_link)
    elif VERBOSE:
        print("Link to directory block OK, {!s}".format(next))

    if not QUIET:
        print("Disk name: {}   Disk id: {}".format(IMAGE.name.decode(), IMAGE.id.decode()))
    if errors == 0 and not QUIET:
        print("OK")

    return errors


def check_bam():
    """Check Block Availability Map."""
    errors = 0

    if not QUIET:
        print("\nChecking Block Availability Map...")
    for track in range(1, IMAGE.MAX_TRACK+1):
        total, bits = IMAGE.bam.get_entry(track)
        counted = bits.count('1')
        if total != counted:
            msg = "Mismatch in track {:d} total and bits, {:d} & {:d} ({})".format(track, total, counted, bits)
            errors += fix_error(msg, fix_bam_entry, track=track)
        if VERBOSE:
            print("Track: {:2d}   Free blocks: {:2d}   Free bits: {}".format(track, total, bits))

    if not IMAGE.bam.is_allocated(IMAGE.DIR_TRACK, 0):
        msg = "Track {:d}:0 not allocated".format(IMAGE.DIR_TRACK)
        errors += fix_error(msg, fix_unalloc_block, block=Block(IMAGE, IMAGE.DIR_TRACK, 0))
    elif VERBOSE:
        print("Block {:d}:0 allocated".format(IMAGE.DIR_TRACK))
    remember_used(IMAGE.DIR_TRACK, 0)

    extra_track = None
    if isinstance(IMAGE, D71Image):
        # all of track 53 should be marked as allocated
        extra_track = IMAGE.EXTRA_BAM_TRACK
        extra_sectors = range(0, IMAGE.max_sectors(IMAGE.EXTRA_BAM_TRACK))
    elif isinstance(IMAGE, D81Image):
        # BAM sectors 1 and 2 should be marked as allocated
        extra_track = IMAGE.DIR_TRACK
        extra_sectors = range(1, 3)

    if extra_track is not None:
        for sector in extra_sectors:
            if not IMAGE.bam.is_allocated(extra_track, sector):
                msg = "Track {:d}:{:d} not allocated".format(extra_track, sector)
                errors += fix_error(msg, fix_unalloc_block, block=Block(IMAGE, extra_track, sector))
            elif VERBOSE:
                print("Block {:d}:{:d} allocated".format(extra_track, sector))
            remember_used(extra_track, sector)

    if errors == 0 and not QUIET:
        print("OK")

    return errors


def check_dir_links():
    """Check the chain of blocks in the directory."""
    errors = 0
    block = Block(IMAGE, IMAGE.DIR_TRACK, IMAGE.DIR_SECTOR)

    while True:
        if not IMAGE.bam.is_allocated(block.track, block.sector):
            msg = "Block {!s} not allocated".format(block)
            errors += fix_error(msg, fix_unalloc_block, block=block)
        elif VERBOSE:
            print("Block {!s} allocated".format(block))
        remember_used(block.track, block.sector)
        if block.is_final:
            if block.data_size != 0xfe:
                msg = "Block {!s} has invalid data size, {:d}".format(block, block.data_size)
                errors += fix_error(msg, fix_data_size, block=block, size=0xfe)
            elif VERBOSE:
                print("Block {!s} checked".format(block))
            # end of chain
            break
        if VERBOSE:
            print("Block {!s} checked".format(block))
        try:
            next_block = block.next_block()
            if next_block.sector in USED_BLOCKS[next_block.track]:
                msg = "Block {!s} links to previous directory block {!s}".format(block, next_block)
                # truncate directory chain
                errors += fix_error(msg, fix_data_size, block=block, size=0xfe)
                break
            block = next_block
        except ValueError:
            ts = block.get(0, 2)
            msg = "Block {!s} has invalid link, {:d}:{:d}".format(block, *ts)
            # truncate directory chain
            errors += fix_error(msg, fix_data_size, block=block, size=0xfe)
            break

    return errors


def check_directory():
    """Check directory integrity."""

    if not QUIET:
        print("\nChecking directory...")
    # first check the integrity of the linked directory blocks
    errors = check_dir_links()

    # next check the basic integrity of each directory entry
    entry = 1
    for path in IMAGE.iterdir():
        raw_ftype = path.entry._file_type()
        if raw_ftype & 7 > path.entry.max_file_type():
            msg = "Entry {:2d}, invalid file type, ${:02x}".format(entry, raw_ftype)
            errors += fix_error(msg, fix_ftype, entry=path.entry, ftype='PRG')
        elif VERBOSE:
            print("Entry {:2d} has valid file type, ${:02x} ({})".format(entry, raw_ftype, path.entry.file_type))

        try:
            first_block = path.entry.first_block()
            if VERBOSE:
                print("Entry {:2d}, link to first block OK, {!s}".format(entry, first_block))
        except ValueError:
            # invalid track/sector
            ts = path.entry.start_ts
            msg = "Entry {:2d}, invalid link to first data block, {:d}:{:d}".format(entry, *ts)
            # missing file contents, delete the entry
            errors += fix_error(msg, fix_ftype, entry=path.entry, ftype=0)

        ss_track, ss_sector = path.entry.side_sector_ts
        if path.entry.file_type == 'REL':
            if path.entry.record_len == 0:
                msg = "Entry {:2d}, invalid relative file record length, {:d}".format(entry, path.entry.record_len)
                # missing relative file info, delete the entry
                errors += fix_error(msg, fix_ftype, entry=path.entry, ftype=0)
            elif VERBOSE:
                print("Entry {:2d}, relative file record length {:d}".format(entry, path.entry.record_len))
        elif path.entry.record_len or ss_track or ss_sector:
            msg = "Entry {:2d}, spurious relative file data, {:d}:{:d} {:d}".format(entry, ss_track, ss_sector, path.entry.record_len)
            errors += fix_error(msg, fix_rel_data, entry=path.entry)
        entry += 1

    if errors == 0 and not QUIET:
        print("OK")

    return errors


def check_side_sectors(path, file_blocks):
    """Check the integrity of side sectors of a relative file."""
    errors = 0
    sector_count = 0
    all_side_sectors = []

    if VERBOSE:
        print("Checking side sectors of relative file {!s}".format(path.name))
    # first check the integrity of the linked side sectors
    rebuild = False
    try:
        side_sector = SideSector(IMAGE, *path.entry.side_sector_ts)
    except ValueError:
        ss_track, ss_sector = path.entry.side_sector_ts
        print("File {!s}, invalid link to side sector, {:d}:{:d}".format(path.name, ss_track, ss_sector))
        rebuild = True
    if not rebuild and side_sector.sector in USED_BLOCKS[side_sector.track]:
        print("File {!s}, first side sector {!s} is in use".format(path.name, side_sector))
        rebuild = True

    while not rebuild and side_sector:
        # this file block usage
        all_side_sectors.append(side_sector)

        try:
            next_ss = side_sector.next_block()
        except ValueError:
            # invalid track/sector
            ts = side_sector.get(0, 2)
            print("Side sector {:d}, invalid link to side sector, {:d}:{:d}".format(sector_count, *ts))
            rebuild = True
            break
        if next_ss:
            if next_ss in all_side_sectors:
                print("Side sector {:d}, links to previous side sector {!s}".format(sector_count, next_ss))
                rebuild = True
                break
            if next_ss.sector in USED_BLOCKS[next_ss.track]:
                print("Side sector {:d}, links to in use block {!s}".format(sector_count, next_ss))
                rebuild = True
                break
            if sector_count+1 == SideSector.MAX_SIDE_SECTORS:
                print("Side sector count exceeds maximum ({:d})".format(SideSector.MAX_SIDE_SECTORS))
                rebuild = True
                break
        if VERBOSE:
            print("link to side sector {:d} OK, {!s}".format(sector_count, side_sector))
        side_sector = next_ss
        sector_count += 1

    if rebuild:
        msg = "Side sector chain needs rebuilding"
        errors = fix_error(msg, fix_rebuild_ss_chain, entry=path.entry, file_blocks=file_blocks)
        side_sectors = [] if errors else SideSector(IMAGE, *path.entry.side_sector_ts).all_side_sectors()
        # either the chain will have been re-created or it is still inconsistent
        return errors, len(side_sectors)

    sector_count = 0
    side_sector = SideSector(IMAGE, *path.entry.side_sector_ts)

    while side_sector:
        # whole image block usage
        remember_used(side_sector.track, side_sector.sector)

        if side_sector.number != sector_count:
            msg = "Side sector {:d}, mismatch in index, {:d}".format(sector_count, side_sector.number)
            errors += fix_error(msg, fix_ss_number, side_sector=side_sector, number=sector_count)
        elif VERBOSE:
            print("Side sector {:d} index OK".format(sector_count))
        if side_sector.record_len != path.entry.record_len:
            msg = "Side sector {:d}, mismatch in record length, {:d} & {:d} (directory)".format(sector_count, side_sector.record_len, path.entry.record_len)
            errors += fix_error(msg, fix_ss_rec_len, side_sector=side_sector, record_len=path.entry.record_len)
        elif VERBOSE:
            print("Side sector {:d}, side sector record length matches directory".format(sector_count))
        if not IMAGE.bam.is_allocated(side_sector.track, side_sector.sector):
            msg = "Side sector {:d} not allocated".format(sector_count)
            errors += fix_error(msg, fix_unalloc_block, block=side_sector)
        elif VERBOSE:
            print("Side sector {:d} allocated".format(sector_count))

        # next check the peer side sector links are correct
        this_ss_list = side_sector.all_side_sectors()
        if this_ss_list != [(ss.track, ss.sector) for ss in all_side_sectors]:
            this_ss_str = ', '.join(["{}:{}".format(t, s) for t, s in this_ss_list])
            all_ss_str = ', '.join([str(ss) for ss in all_side_sectors])
            msg = "Side sector {:d}, mismatch in peer links, {} & {}".format(sector_count, this_ss_str, all_ss_str)
            errors += fix_error(msg, fix_ss_list, side_sector=side_sector, ss_list=all_side_sectors)
        elif VERBOSE:
            print("Side sector {:d} peer links OK".format(sector_count))

        # finally check the data block links are correct
        this_data_list = side_sector.all_data_blocks()
        file_slice = file_blocks[sector_count*SideSector.MAX_DATA_LINKS:(sector_count+1)*SideSector.MAX_DATA_LINKS]
        if this_data_list != [(d.track, d.sector) for d in file_slice]:
            this_data_str = ', '.join(["{}:{}".format(t, s) for t, s in this_data_list])
            slice_str = ', '.join([str(d) for d in file_slice])
            msg = "Side sector {:d}, mismatch in data links, {} & {}".format(sector_count, this_data_str, slice_str)
            errors += fix_error(msg, fix_ss_data, side_sector=side_sector, data_list=file_slice)
        elif VERBOSE:
            print("Side sector {:d} contains {:d} data links".format(sector_count, len(this_data_list)))

        side_sector = side_sector.next_block()
        sector_count += 1

    return errors, sector_count


def check_path(path):
    """Check integrity of the contents of a file."""
    errors = 0
    file_blocks = []

    try:
        block = path.entry.first_block()
    except ValueError:
        print("File {!s}, invalid link to first data block (unfixed)".format(path.name))
        return 1
    if block.sector in USED_BLOCKS[block.track]:
        # linked to another file
        msg = "File {!s}, first block {!s} is in use".format(path.name, block)
        e = fix_error(msg, fix_entry_linked, entry=path.entry)
        if e:
            # not fixed, skip further processing of this file
            return 1
        block = path.entry.first_block()

    while block:
        # whole image block usage
        remember_used(block.track, block.sector)
        # this file block usage
        file_blocks.append(block)
        if VERBOSE:
            print("File {!s}, link to block OK, {!s}".format(path.name, block))
        if not IMAGE.bam.is_allocated(block.track, block.sector):
            msg = "File {!s}, block {!s} not allocated".format(path.name, block)
            errors += fix_error(msg, fix_unalloc_block, block=block)
        elif VERBOSE:
            print("File {!s}, block {!s} allocated".format(path.name, block))

        try:
            next_block = block.next_block()
        except ValueError:
            # invalid track/sector
            ts = block.get(0, 2)
            msg = "File {!s}, invalid link to block, {:d}:{:d}".format(path.name, *ts)
            # truncate file
            errors += fix_error(msg, fix_data_size, block=block, size=0xfe)
            break
        if next_block:
            if next_block in file_blocks:
                msg = "File {!s}, block {!s} links to previous data block {!s}".format(path.name, block, next_block)
                # truncate file
                errors += fix_error(msg, fix_data_size, block=block, size=0xfe)
                break
            if next_block.sector in USED_BLOCKS[next_block.track]:
                # linked to another file
                msg = "File {!s}, block {!s} links to in use block {!s}".format(path.name, block, next_block)
                errors += fix_error(msg, fix_file_linked, block=block)
                next_block = block.next_block()
        else:
            try:
                block_size = block.data_size
                if VERBOSE:
                    print("File {!s}, final block contains {:d} bytes".format(path.name, block_size))
            except IndexError:
                msg = "File {!s}, final block has invalid size".format(path.name)
                errors += fix_error(msg, fix_data_size, block=block, size=0xfe)

        block = next_block

    blocks_used = len(file_blocks)
    if path.entry.file_type == 'REL':
        ss_errors, ss_blocks_used = check_side_sectors(path, file_blocks)
        errors += ss_errors
        blocks_used += ss_blocks_used

    if blocks_used != path.size_blocks:
        msg = "File {!s}, mismatch in blocks used, {:d} & {:d} (actual)".format(path.name, path.size_blocks, blocks_used)
        errors += fix_error(msg, fix_block_count, entry=path.entry, count=blocks_used)
    elif VERBOSE:
        print("File {!s} uses {:d} blocks".format(path.name, blocks_used))

    return errors


def check_partition(path):
    """Check integrity of partition."""
    errors = 0

    try:
        block = path.entry.first_block()
    except ValueError:
        print("Partition {!s}, invalid link to first data block (unfixed)".format(path.name))
        return 1

    block_count = path.size_blocks
    while block_count:
        block_count -= 1

        if block.sector in USED_BLOCKS[block.track]:
            # linked to another file
            print("Partition {!s}, block {!s} is in use (unfixed)".format(path.name, block))
            return 1

        # whole image block usage
        remember_used(block.track, block.sector)
        if not IMAGE.bam.is_allocated(block.track, block.sector):
            msg = "Partition {!s}, block {!s} not allocated".format(path.name, block)
            errors += fix_error(msg, fix_unalloc_block, block=block)
        elif VERBOSE:
            print("Partition {!s}, block {!s} allocated".format(path.name, block))

        if block_count:
            # blocks are allocated contiguously
            try:
                sector = block.sector+1
                if sector >= IMAGE.max_sectors(block.track):
                    track = block.track+1
                    sector = 0
                else:
                    track = block.track
                block = Block(IMAGE, track, sector)
            except ValueError:
                print("Partition {!s}, invalid block, {:d}:{:d}".format(path.name, track, sector))
                return 1

    if VERBOSE:
        print("Partition {!s} uses {:d} blocks".format(path.name, path.size_blocks))

    return errors


def check_files():
    """Check integrity of all files."""
    errors = 0

    if not QUIET:
        print("\nChecking files...")
    for path in IMAGE.iterdir():
        if not path.entry.closed:
            msg = "File {!s}, unclosed".format(path.name)
            errors += fix_error(msg, fix_unclosed, entry=path.entry)

        if path.entry.file_type == 'CBM':
            errors += check_partition(path)
        else:
            errors += check_path(path)

    if errors == 0 and not QUIET:
        print("OK")

    return errors


def check_allocation():
    global USED_BLOCKS
    errors = 0

    if not QUIET:
        print("\nChecking block allocation...")
    for track in range(1, IMAGE.MAX_TRACK+1):
        max_sectors = IMAGE.max_sectors(track)
        _, bits = IMAGE.bam.get_entry(track)
        bam_used = {i for i, b in enumerate(bits) if b == '0' and i < max_sectors}
        delta = bam_used-USED_BLOCKS[track]

        if delta:
            delta_s = ', '.join([str(b) for b in delta])
            msg = "Track {:d}, sectors {} marked allocated when unused".format(track, delta_s)
            # generate an updated bitmask for sectors actually used
            fixed_bits = ''.join(['1' if i in delta else b for i, b in enumerate(bits)])
            errors += fix_error(msg, fix_track_alloc, track=track, bits=fixed_bits)
        elif VERBOSE:
            print("Track {:2d} OK".format(track))

    if errors == 0 and not QUIET:
        print("OK")

    return errors


def check_image(image_path):
    """Check the integrity of an image, return the number of uncorrected errors."""
    global IMAGE
    global USED_BLOCKS

    USED_BLOCKS = defaultdict(set)

    mode = 'w' if FIX else 'r'
    with DiskImage(image_path, mode) as IMAGE:
        errors = check_misc()
        errors += check_bam()
        errors += check_directory()
        errors += check_files()
        errors += check_allocation()

        if VERBOSE:
            print()
            for line in IMAGE.directory('ascii'):
                print(line)

    return errors


def fix_dos_version():
    """Fix DOS version field."""
    IMAGE.dos_version = IMAGE.DOS_VERSION
    if VERBOSE:
        print("Setting DOS version to "+chr(IMAGE.DOS_VERSION))
    return 0


def fix_dos_types():
    """Fix DOS format type fields."""
    IMAGE.dos_type = IMAGE.DOS_FORMAT
    if VERBOSE:
        print("Setting DOS format to "+chr(IMAGE.DOS_FORMAT))
    return 0


def fix_dir_link():
    """Fix link to directory block."""
    dir_block = Block(IMAGE, IMAGE.DIR_TRACK, 1)
    block = Block(IMAGE, IMAGE.DIR_TRACK, 0)
    block.set_next_block(dir_block)
    if VERBOSE:
        print("Setting link to {!s}".format(dir_block))
    return 0


def fix_bam_entry(track):
    """Fix track entry in BAM."""
    _, bits = IMAGE.bam.get_entry(track)
    counted = bits.count('1')
    IMAGE.bam.set_entry(track, counted, bits)
    if VERBOSE:
        print("Setting track {:d} to {:d} & {}".format(track, counted, bits))
    return 0


def fix_unalloc_block(block):
    """Allocate an in-use block."""
    IMAGE.bam.set_allocated(block.track, block.sector)
    if VERBOSE:
        print("Allocating block {!s}".format(block))
    return 0


def fix_data_size(block, size):
    """Fix data used in a block."""
    block.data_size = size
    if VERBOSE:
        print("Setting data size of {!s} to {:d}".format(block, size))
    return 0


def fix_ftype(entry, ftype):
    """Fix entry file type."""
    entry.file_type = ftype
    if VERBOSE:
        print("Setting entry file type to "+entry.file_type)
    return 0


def fix_rel_data(entry):
    """Clear relative file fields."""
    entry.side_sector_ts = (0, 0)
    entry.record_len = 0
    if VERBOSE:
        print("Clearing relative file data")
    return 0


def fix_ss_number(side_sector, number):
    """Fix side sector number."""
    side_sector.number = number
    if VERBOSE:
        print("Setting side sector index to {:d}".format(number))
    return 0


def fix_ss_rec_len(side_sector, record_len):
    """Fix side sector record length."""
    side_sector.record_len = record_len
    if VERBOSE:
        print("Setting side sector record length to {:d}".format(record_len))
    return 0


def fix_rebuild_ss_chain(entry, file_blocks):
    """Reconstruct all side sectors from scratch."""
    side_sectors = []
    alloc_from = file_blocks[-1]
    last_ss = None

    while file_blocks:
        block = IMAGE.alloc_next_block(alloc_from.track, alloc_from.sector)
        ss = SideSector(IMAGE, block.track, block.sector)
        remember_used(block.track, block.sector)
        ss.number = len(side_sectors)
        ss.record_len = entry.record_len
        ss.set_data_blocks(file_blocks[:SideSector.MAX_DATA_LINKS])
        if last_ss:
            last_ss.set_next_block(ss)
        else:
            entry.side_sector_ts = (ss.track, ss.sector)
        side_sectors.append(ss)
        last_ss = ss
        alloc_from = ss
        file_blocks = file_blocks[SideSector.MAX_DATA_LINKS:]

    # update the peer links of all side sectors
    for ss in side_sectors:
        ss.set_peers([(ss.track, ss.sector) for ss in side_sectors])

    if VERBOSE:
        print("Rebuilding {:d} side sectors".format(len(side_sectors)))
    return 0


def fix_ss_list(side_sector, ss_list):
    """Fix list of peer side sectors."""
    side_sector.set_peers([(ss.track, ss.sector) for ss in ss_list])
    if VERBOSE:
        print("Setting peer links to", ', '.join([str(ss) for ss in ss_list]))
    return 0


def fix_ss_data(side_sector, data_list):
    """Fix list of data blocks in side sector."""
    side_sector.set_data_blocks(data_list)
    if VERBOSE:
        print("Setting data links to", ', '.join([str(d) for d in data_list]))
    return 0


def fix_unclosed(entry):
    """Mark an entry as closed."""
    entry.closed = True
    if VERBOSE:
        print("Setting entry file state as closed")
    return 0


def fix_entry_linked(entry):
    """Clone a file whose first block is linked to another."""
    new_block = IMAGE.clone_chain(entry.first_block())
    entry.start_ts = (new_block.track, new_block.sector)
    if VERBOSE:
        print("Duplicating blocks, new first block {!s}".format(new_block))
    return 0


def fix_file_linked(block):
    """Clone chain of blocks for file linked to another."""
    new_block = IMAGE.clone_chain(block.next_block())
    block.set_next_block(new_block)
    if VERBOSE:
        print("Duplicating blocks, new next block {!s}".format(new_block))
    return 0


def fix_block_count(entry, count):
    """Fix entry size in blocks."""
    entry.size = count
    if VERBOSE:
        print("Setting block count to {:d}".format(count))
    return 0


def fix_track_alloc(track, bits):
    """Fix BAM entry for a track."""
    IMAGE.bam.set_entry(track, bits.count('1'), bits)
    if VERBOSE:
        print("Setting track {:d} bits to {}".format(track, bits))
    return 0


def main():
    global FIX
    global YES
    global VERBOSE
    global QUIET

    parser = argparse.ArgumentParser(description='Check and fix Commodore disk images.')
    parser.add_argument('image', type=Path, help='image filename')
    parser.add_argument('--fix', '-f', action='store_true', help='Fix errors detected')
    parser.add_argument('--yes', '-y', action='store_true', help='Answer questions with "yes"')
    parser.add_argument('--verbose', '-v', action='store_true', help='Verbose output')
    parser.add_argument('--quiet', '-q', action='store_true', help='No output')

    args = parser.parse_args()
    FIX = args.fix
    YES = args.yes
    VERBOSE = args.verbose
    QUIET = args.quiet

    try:
        errors = check_image(args.image)
    except KeyboardInterrupt:
        sys.exit("\nAbort, discarding all changes")
    except Exception as e:
        sys.exit("\nException: %s" % str(e))

    if errors:
        sys.exit("\n{:d} unfixed errors".format(errors))



import sys

import cbmcodecs

from d64 import DiskImage


fname = sys.argv[2].encode('petscii-c64en-uc')

with DiskImage(sys.argv[1], mode='w') as image:
    with image.path(fname).open('w', ftype='rel', record_len=220) as f:
        for n in range(0, 70):
            f.write(b'ONE')
            f.write(b'TWO')
            f.write(b'THREE')

with DiskImage(sys.argv[1]) as image:
    with image.path(fname).open() as f:
        count = 0
        while True:
            record = f.read_record()
            if not record:
                break
            print(count, record)
            count += 1
